-- ------------------------------------------------------------------------------ --
--                                TradeSkillMaster                                --
--                          https://tradeskillmaster.com                          --
--    All Rights Reserved - Detailed license information included with addon.     --
-- ------------------------------------------------------------------------------ --

local TSM = select(2, ...) ---@type TSM
local GoldTracker = TSM.Accounting:NewPackage("GoldTracker") ---@type AddonPackage
local ClientInfo = TSM.LibTSMWoW:Include("Util.ClientInfo")
local Event = TSM.LibTSMWoW:Include("Service.Event")
local DelayTimer = TSM.LibTSMWoW:IncludeClassType("DelayTimer")
local Math = TSM.LibTSMUtil:Include("Lua.Math")
local Log = TSM.LibTSMUtil:Include("Util.Log")
local TempTable = TSM.LibTSMUtil:Include("BaseType.TempTable")
local SessionInfo = TSM.LibTSMWoW:Include("Util.SessionInfo")
local DefaultUI = TSM.LibTSMWoW:Include("UI.DefaultUI")
local GoldLog = TSM.LibTSMTypes:IncludeClassType("GoldLog")
local PlayerInfo = TSM.Include("Service.PlayerInfo")
local private = {
	truncateGoldLog = {},
	characterLog = {}, ---@type table<string,GoldLog>
	characterLastUpdate = {}, ---@type table<string,number>
	guildLog = {}, ---@type table<string,GoldLog>
	guildLastUpdate = {}, ---@type table<string,number>
	currentCharacterKey = nil, ---@type string
	playerLogCount = 0,
	playerGoldRetryTimer = nil,
	settings = nil,
}
local SECONDS_PER_MIN = 60
local SECONDS_PER_DAY = SECONDS_PER_MIN * 60 * 24



-- ============================================================================
-- Module Functions
-- ============================================================================

function GoldTracker.OnInitialize(settingsDB)
	private.playerGoldRetryTimer = DelayTimer.New("PLAYER_GOLD_RETRY", private.PlayerLogGold)
	if ClientInfo.HasFeature(ClientInfo.FEATURES.GUILD_BANK) then
		DefaultUI.RegisterGuildBankVisibleCallback(private.GuildLogGold, true)
		Event.Register("GUILDBANK_UPDATE_MONEY", private.GuildLogGold)
	end
	Event.Register("PLAYER_MONEY", private.PlayerLogGold)

	private.settings = settingsDB:NewView()
		:AddKey("sync", "internalData", "goldLog")
		:AddKey("sync", "internalData", "goldLogLastUpdate")
		:AddKey("sync", "internalData", "money")
		:AddKey("factionrealm", "internalData", "guildGoldLog")
		:AddKey("factionrealm", "internalData", "guildGoldLogLastUpdate")
		:AddKey("factionrealm", "internalData", "characterGuilds")
		:AddKey("global", "coreOptions", "regionWide")

	-- Get a list of known guilds and load character gold log data
	local validGuilds = TempTable.Acquire()
	for _, data, character, factionrealm in private.settings:AccessibleValueIterator("goldLog", not private.settings.regionWide) do
		local guild = private.settings:GetForScopeKey("characterGuilds", factionrealm)[character]
		if guild then
			validGuilds[guild] = true
		end
		local lastUpdate = private.settings:GetForScopeKey("goldLogLastUpdate", character, factionrealm) or 0
		local characterKey = SessionInfo.FormatCharacterName(character, factionrealm, true)
		private.LoadCharacterGoldLog(characterKey, data, lastUpdate)
	end

	-- Load guild gold log data
	for _, guildData, factionrealm in private.settings:AccessibleValueIterator("guildGoldLog", not private.settings.regionWide) do
		for guild, data in pairs(guildData) do
			local log = GoldLog.Load(data) or GoldLog.New()
			private.guildLog[guild] = log
			local endMinute = log:GetEndMinute()
			local lastEntryTime = endMinute and endMinute * SECONDS_PER_MIN or math.huge
			local lastUpdate = private.settings:GetForScopeKey("guildGoldLogLastUpdate", factionrealm)
			if not validGuilds[guild] and max(lastEntryTime, lastUpdate and lastUpdate[guild] or 0) < time() - 30 * SECONDS_PER_DAY then
				-- This guild may not be valid and the last entry is over 30 days old, so truncate the data
				private.truncateGoldLog[guild] = lastEntryTime
			end
		end
	end

	TempTable.Release(validGuilds)
	private.currentCharacterKey = SessionInfo.FormatCharacterName(SessionInfo.GetCharacterName(), SessionInfo.GetFactionrealmName(), true)
	assert(private.characterLog[private.currentCharacterKey])
end

function GoldTracker.OnEnable()
	-- Log the current player gold (need to wait for OnEnable, otherwise GetMoney() returns 0 when first logging in)
	private.PlayerLogGold()
end

function GoldTracker.OnDisable()
	private.PlayerLogGold()
	private.settings.goldLog = private.characterLog[private.currentCharacterKey]:Serialize()
	private.settings.goldLogLastUpdate = private.characterLastUpdate[private.currentCharacterKey]
	local guild = PlayerInfo.GetPlayerGuild(SessionInfo.GetCharacterName(), SessionInfo.GetFactionrealmName())
	if guild and private.guildLog[guild] then
		private.settings.guildGoldLog[guild] = private.guildLog[guild]:Serialize()
		private.settings.guildGoldLogLastUpdate[guild] = private.guildLastUpdate[guild]
	end
end

function GoldTracker.CharacterGuildIterator()
	return private.CharacterGuildIteratorHelper
end

function GoldTracker.GetGoldAtTime(timestamp, ignoredCharactersGuilds)
	local value = 0
	local timestampMinute = floor(timestamp / SECONDS_PER_MIN)
	for key, log in private.LogIterator(ignoredCharactersGuilds) do
		if (private.truncateGoldLog[key] or math.huge) > timestamp then
			value = value + log:GetValue(timestampMinute)
		end
	end
	return value
end

function GoldTracker.GetGraphTimeRange(ignoredCharactersGuilds)
	local minTime = Math.Floor(time(), SECONDS_PER_MIN)
	for _, log in private.LogIterator(ignoredCharactersGuilds) do
		local startMinute = log:GetStartMinute()
		if startMinute then
			minTime = min(minTime, startMinute * SECONDS_PER_MIN)
		end
	end
	return minTime, Math.Floor(time(), SECONDS_PER_MIN), SECONDS_PER_MIN
end



-- ============================================================================
-- Private Helper Functions
-- ============================================================================

function private.LoadCharacterGoldLog(characterKey, data, lastUpdate)
	assert(not private.characterLog[characterKey])
	local log = GoldLog.Load(data)
	if not log then
		Log.Err("Failed to decode (%s, %d)", characterKey, #data)
		private.characterLog[characterKey] = GoldLog.New()
		return
	end
	log:Clean()
	private.characterLastUpdate[characterKey] = lastUpdate
	private.characterLog[characterKey] = log
end

function private.GuildLogGold()
	local guild = GetGuildInfo("player")
	local isGuildLeader = IsGuildLeader()
	if guild and not isGuildLeader then
		-- check if our alt is the guild leader
		for i = 1, GetNumGuildMembers() do
			local name, _, rankIndex = GetGuildRosterInfo(i)
			if name and rankIndex == 0 and PlayerInfo.IsPlayer(gsub(name, "%-", " - "), true) then
				isGuildLeader = true
			end
		end
	end
	if guild and isGuildLeader then
		if not private.guildLog[guild] then
			private.guildLog[guild] = GoldLog.New()
		end
		private.UpdateGuildGoldLog(guild, GetGuildBankMoney())
	end
end

function private.PlayerLogGold()
	-- GetMoney sometimes returns 0 for a while after login, so keep trying for 30 seconds before recording a 0
	local money = GetMoney()
	if money == 0 and private.playerLogCount < 30 then
		private.playerLogCount = private.playerLogCount + 1
		private.playerGoldRetryTimer:RunForTime(1)
		return
	end
	private.playerLogCount = 0
	private.UpdateCurrentCharacterGoldLog(money)
	private.settings.money = money
end

function private.UpdateGuildGoldLog(guild, copper)
	local currentMinute = floor(time() / SECONDS_PER_MIN)
	private.guildLog[guild]:Append(currentMinute, private.RoundCopperValue(copper))
	private.guildLastUpdate[guild] = time()
end

function private.UpdateCurrentCharacterGoldLog(copper)
	local currentMinute = floor(time() / SECONDS_PER_MIN)
	private.characterLog[private.currentCharacterKey]:Append(currentMinute, private.RoundCopperValue(copper))
	private.characterLastUpdate[private.currentCharacterKey] = time()
end

function private.RoundCopperValue(copper)
	return Math.Round(copper, COPPER_PER_GOLD * ((ClientInfo.IsRetail() and 1000) or (ClientInfo.IsVanillaClassic() and 1) or 100))
end

function private.CharacterGuildIteratorHelper(_, lastKey)
	local key, isGuild = nil, nil
	if not lastKey or private.characterLog[lastKey] then
		key = next(private.characterLog, lastKey)
		isGuild = false
		if not key then
			lastKey = nil
		end
	end
	if not key then
		key = next(private.guildLog, lastKey)
		isGuild = true
	end
	if not key then
		return
	end
	return key, isGuild
end

function private.LogIterator(ignoredCharactersGuilds)
	return private.LogIteratorHelper, ignoredCharactersGuilds
end

function private.LogIteratorHelper(ignoredCharactersGuilds, lastKey)
	while true do
		local key, isGuild = private.CharacterGuildIteratorHelper(nil, lastKey)
		if not key then
			return
		elseif not ignoredCharactersGuilds[key] then
			if isGuild then
				return key, private.guildLog[key]
			else
				return key, private.characterLog[key]
			end
		end
	end
end
